from paraminjector.typed_callable import TypedCallable


class FunctionSignatureInvalid(Exception):
    def __init__(self, func: TypedCallable, cause: str) -> None:
        super().__init__(f"{cause.rstrip('.')},\ncaused by {func.description}.")



__version__ = "0.1.0"

from .static import ParamInjector
from .dynamic import call_with_args_async, call_with_args, call_with_args_maybe_async
from .typed_callable import TypedCallable

__all__ = ["ParamInjector", "call_with_args_async", "call_with_args", "call_with_args_maybe_async"]



import inspect
import traceback

import itertools
from inspect import _empty
from typing import Any, Tuple, Type, Union, Dict, cast

from paraminjector.typed_callable import TypedCallable
from paraminjector.exceptions import FunctionSignatureInvalid


def analyze_signature(
    func: TypedCallable,
    available_args: Dict[Type, object],
    fixed_pos_args: Tuple[Any] = None,
    follow_wrapped: bool = True,
) -> Dict[str, object]:
    if fixed_pos_args is not None:
        params_to_inject = get_detected_params(func, available_args, fixed_pos_args)
    else:
        params_to_inject = func.signature.parameters

    result: Dict[str, object] = dict()

    for param_name, parameter in params_to_inject.items():
        parameter = cast(inspect.Parameter, parameter)

        # account for union type
        annotated_types = (
            type_args
            if (type_args := getattr(parameter.annotation, "__args__", None))
            else parameter.annotation
        )

        # TODO: add try catch for "TypeError: Subscripted generics cannot be used with class and instance checks"
        possible_targets = [a for a in available_args if issubclass(a, annotated_types)]

        if len(possible_targets) > 1:

            # TODO: this is gonna happen

            raise ValueError(
                f"Found more than one possible target for parameter {parameter.__repr__()}:"
                f" {possible_targets}"
            )
        elif not possible_targets:
            if parameter.default is not inspect._empty:
                # parameter has default value but we cannot determine it
                continue
            else:
                raise ValueError(
                    f"Parameter {parameter.__repr__()} cannot be filled. TODO: more info"
                )

        target = possible_targets[0]

        param = _find_contravariant_argument(available_args, target)

        if param is not inspect._empty:
            result[param_name] = param
        else:
            if parameter.default == inspect.Parameter.empty:
                raise ValueError("Non-default parameter in callback signature whose type blabla")
            else:
                pass  # ignore

    return result


def get_detected_params(
    func: TypedCallable, available_args: Dict[Type, object], fixed_pos_args: Tuple[Any]
) -> Dict[str, inspect.Parameter]:

    n_fixed = len(fixed_pos_args)
    n_declared = len(func.signature.parameters)

    if n_fixed > n_declared:
        raise FunctionSignatureInvalid(func, f"Expected at least {n_fixed} parameters")

    param_kv_pairs: Dict[int, Tuple[str, inspect.Parameter]] = {
        n: kv for n, kv in enumerate(func.signature.parameters.items())
    }

    # If it has a type annotation, make sure we match that with the arguments we're going to inject.
    for n, pos_arg in enumerate(fixed_pos_args):
        name, param = param_kv_pairs[n]

        # TODO: How can pos_arg be None? Write test!

        try:
            if param.annotation and not issubclass(type(pos_arg), param.annotation):
                raise FunctionSignatureInvalid(
                    func,
                    f"Argument {n} to {func.name} should always be a supertype of "
                    f"{type(pos_arg)}.",
                )
        except TypeError as e:
            traceback.print_exc()

    try:
        return {
            k: v
            for k, v in itertools.islice(func.signature.parameters.items(), n_fixed, n_declared)
        }
    except KeyError:
        raise FunctionSignatureInvalid("lala TODO")


def _find_contravariant_argument(
    refs: Dict[Type, object], param_type: Type
) -> Union[_empty, object]:
    exact_match = refs.get(param_type, None)
    if exact_match is not None:
        # Exact type match
        return exact_match

    covariant_match = next((p for p in refs if issubclass(p, param_type)), _empty)
    return refs[covariant_match]



import inspect
from typing import (
    Any,
    Callable,
    ClassVar,
    Generic,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    AbstractSet,
    Hashable,
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Sequence,
    AsyncIterator,
    AsyncIterable,
    Coroutine,
    Collection,
    AsyncGenerator,
    Deque,
    Dict,
    List,
    Set,
    FrozenSet,
    NamedTuple,
    Generator,
    cast,
    overload,
    TYPE_CHECKING,
)
from pprint import pprint

from typing import Callable, Type, Union, Dict, cast

from paraminjector.typed_callable import TypedCallable
from paraminjector._injector import _find_contravariant_argument, analyze_signature


class ParamInjector:
    available_references: Dict[Type, object]

    def __init__(
        self,
        func: Callable,
        available_args: Dict[Type, object],
        fixed_pos_arg_types: Dict[Type, object] = None,
        follow_wrapped: bool = True,
    ):
        """
        If you expect the callback to change during runtime, use dynamic injections instead.

        Args:
            func: The user-defined function to call. It must not change during runtime.
            available_args:
        """
        self.func = func
        analyze_signature(
            TypedCallable(func),
            available_args=available_args,
            fixed_pos_args=fixed_pos_args,
            follow_wrapped=follow_wrapped,
        )
        self.available_references = available_args

    def map_kwargs(self, callback: Callable, follow_wrapped: bool = True) -> Dict[str, object]:
        sig = inspect.signature(callback, follow_wrapped=follow_wrapped)

        # pprint(sig.parameters)

        result: Dict[str, object] = dict()

        for param_name, parameter in sig.parameters.items():
            print(parameter.default)
            parameter = cast(inspect.Parameter, parameter)

            # TODO: add try catch for "TypeError: Subscripted generics cannot be used with class and instance checks"
            possible_targets = [
                a for a in self.available_references if issubclass(a, parameter.annotation)
            ]

            if len(possible_targets) > 1:
                raise ValueError(
                    f"More than one possible target for parameter {parameter.__repr__()}:"
                    f" {possible_targets}"
                )
            elif not possible_targets:
                if parameter.default is not inspect._empty:
                    # parameter has default value but we cannot determine it
                    continue
                else:
                    raise ValueError(
                        f"Parameter {parameter.__repr__()} cannot be filled. TODO: more info"
                    )

            target = possible_targets[0]

            param = _find_contravariant_argument(self.available_references, parameter.annotation)
            print(param)

            if param is not inspect._empty:
                result[param_name] = param
            else:
                if parameter.default == inspect.Parameter.empty:
                    raise ValueError(
                        "Non-default parameter in callback signature whose type blabla"
                    )
                else:
                    pass  # ignore

            pprint(result)

        return result

    def call(self, *fixed_args, **fixed_kwargs):
        pass

    async def call_async(self, *fixed_args, **fixed_kwargs):
        pass

        # if not used_arg_types:
        #     raise ValueError(
        #         f"No type hints specified for callback '{callback.__name__}'."
        #     )
        #
        # pyrogram_types = {Message, CallbackQuery, InlineQuery, Poll, User}
        #
        # if Update in used_arg_types:
        #     found_arg_type = Update
        # else:
        #     found_arg_type = None
        #
        # for arg in used_arg_types:
        #     if issubclass(arg, Client):
        #         pass
        #     if any((t for t in pyrogram_types if issubclass(arg, t))):
        #         if found_arg_type is not None:
        #             raise ValueError(
        #                 "More than one possible update type found, cannot determine the kind of callback."
        #             )
        #         else:
        #             found_arg_type = arg
        #
        # if not found_arg_type:
        #     raise ValueError(
        #         f"No matching callback type found for signature {type_hints}"
        #     )
        #
        # kwargs = dict(callback=callback, filters=filters)
        # if issubclass(found_arg_type, Message):
        #     return MessageHandler(**kwargs)
        # elif issubclass(found_arg_type, CallbackQuery):
        #     return CallbackQueryHandler(**kwargs)
        # elif issubclass(found_arg_type, InlineQuery):
        #     return InlineQueryHandler(**kwargs)
        # elif issubclass(found_arg_type, Poll):
        #     return PollHandler(**kwargs)
        # elif issubclass(found_arg_type, User):
        #     return UserStatusHandler(**kwargs)
        # elif found_arg_type is Update:
        #     return RawUpdateHandler(callback=callback)
        # else:
        #     raise ValueError(
        #         f"Could not find a matching Pyrogram callback class for type {found_arg_type}."
        #     )



import inspect
import warnings
from dataclasses import dataclass

from cached_property import cached_property
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    Protocol,
    TypeVar,
    Union,
    get_type_hints,
)


T = TypeVar("T", bound=Callable)


@dataclass(frozen=True)
class TypedCallable(Generic[T]):
    func: T

    @cached_property
    def signature(self) -> inspect.Signature:
        return inspect.signature(self.func)

    @cached_property
    def is_coroutine(self) -> bool:
        result = inspect.iscoroutinefunction(self.func)
        if not result and inspect.isawaitable(result):
            # Generally, we don't want to have anything to do with asyncio futures, tasks, and the like.
            warnings.warn(
                f"Callable {self.name} is a {type(self.func)}, which is awaitable, but not a coroutine (async "
                f"def). It is possible that you will always get the same result when it is awaited."
            )
            return True
        return result

    @cached_property
    def type_hints(self) -> Dict[str, Any]:
        return get_type_hints(self.func)

    @cached_property
    def num_parameters(self) -> int:
        return len(self.signature.parameters)

    @cached_property
    def num_non_optional_params(self) -> int:
        return sum((1 for p in self.signature.parameters.values() if p.default is p.empty))

    @cached_property
    def description(self) -> str:
        parts = []
        if inspect.iscoroutinefunction(self.func):
            parts.append("asynchronous")

        # TODO: Could use some more spice for things like classmethods, properties, etc.

        parts.append("method" if inspect.ismethod(self.func) else "function")
        parts.append(self.name)
        parts.append("with signature")
        parts.append(str(self.signature))

        result = " ".join(parts)
        return result[0].upper() + result[1:]

    @property
    def name(self) -> str:
        return self.func.__name__

    def __call__(self, *args, **kwargs) -> Any:
        return self.func(*args, **kwargs)

    def __str__(self) -> str:
        return str(self.func)



import inspect
from typing import Any, Callable, Type, TypeVar, Union, Dict, Tuple
from typing import Awaitable

from paraminjector.typed_callable import TypedCallable
from paraminjector._injector import analyze_signature


def call_with_args(
    func: Callable,
    available_args: Dict[Type, object],
    fixed_pos_args: Tuple = None,
    follow_wrapped: bool = True,  # TODO: implement
) -> Any:
    kwargs = analyze_signature(
        func=TypedCallable(func),
        available_args=available_args,
        fixed_pos_args=fixed_pos_args,
        follow_wrapped=follow_wrapped,
    )

    if fixed_pos_args:
        return func(*fixed_pos_args, **kwargs)
    else:
        return func(**kwargs)


async def call_with_args_maybe_async(
    func: Callable,
    available_args: Dict[Type, object],
    fixed_pos_args: Tuple = None,
    follow_wrapped: bool = True,  # TODO: implement
) -> Any:
    result = call_with_args(
        func, available_args, fixed_pos_args=fixed_pos_args, follow_wrapped=follow_wrapped
    )
    if inspect.isawaitable(result):
        return await result
    return result


async def call_with_args_async(
    func: Callable[[], Awaitable],
    available_args: Dict[Type, object],
    fixed_pos_args: Tuple = None,
    follow_wrapped: bool = True,  # TODO: implement
) -> Any:
    return await call_with_args(
        func, available_args, fixed_pos_args=fixed_pos_args, follow_wrapped=follow_wrapped
    )
